from __future__ import annotations

import json
import re
from dataclasses import MISSING, Field, asdict
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Literal, Type, TypeVar, Union

T = TypeVar("T")
JsonValue = Union[dict, list, str, int, float, bool, Literal[None]]


class Encoder:
    def __init__(self, obj_type: Type[T], encoder: Callable[[T], JsonValue]):
        self.obj_type = obj_type
        self.encoder = encoder

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(obj_type={self.obj_type.__name__})"

    def check(self, obj: Any) -> bool:
        """Check if object can be encoded"""
        return self.obj_type == obj.__class__

    def encode(self, obj: Any) -> dict[str, JsonValue]:
        """Encodes the python object into a JSON object"""
        return {
            "__type__": self.obj_type.__name__,
            "value": self.encoder(obj),
        }


class Decoder:
    def __init__(self, obj_type: Type[T], decoder: Callable[[JsonValue], T]):
        self.obj_type = obj_type
        self.decoder = decoder

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(obj_type={self.obj_type.__name__})"

    def check(self, obj: dict[str, JsonValue]) -> bool:
        """Check if object can be decoded"""
        return self.obj_type.__name__ == obj["__type__"]

    def decode(self, obj: dict[str, JsonValue]) -> Any:
        """Decodes the JSON object into a python object"""
        return self.decoder(obj["value"])


class TableEncoder(Encoder):
    def __init__(self):
        super().__init__(Table, asdict)

    def __hash__(self) -> int:
        return hash(Table)

    def check(self, obj: Any) -> bool:
        """Check if object can be encoded"""
        return isinstance(obj, Table)

    def encode(self, obj: Any) -> dict[str, JsonValue]:
        """Encodes the python object into a JSON object"""
        return {
            "__type__": obj.__class__.__name__,
            "value": self.encoder(obj),
        }


class TableDecoder(Decoder):
    def __init__(self):
        super().__init__(Table, lambda x: None)

    def __hash__(self) -> int:
        return hash(Table)

    @staticmethod
    def sub_types(_type: type) -> dict[str, type]:
        sub_types = {_type.__name__: _type}
        for sub in _type.__subclasses__():
            sub_types.update(TableDecoder.sub_types(sub))
        return sub_types

    def check(self, obj: dict[str, JsonValue]) -> bool:
        """Check if object can be decoded"""
        return obj["__type__"] in TableDecoder.sub_types(Table)

    def decode(self, obj: dict[str, JsonValue]) -> Any:
        """Decodes the JSON object into a python object"""
        return TableDecoder.sub_types(Table)[obj["__type__"]](**obj["value"])


class Comparison:
    def __init__(self, compare: Callable[[Table], bool]):
        self.compare = compare

    def __and__(self, other: Comparison):
        return Comparison(lambda e: self.compare(e) and other.compare(e))

    def __or__(self, other: Comparison):
        return Comparison(lambda e: self.compare(e) or other.compare(e))


class TableField(Field):
    # noinspection PyMissingConstructor
    def __init__(self, field: Field):
        for attr in Field.__slots__:
            setattr(self, attr, getattr(field, attr))

    @property
    def get_default(self) -> Any:
        if self.default_factory is MISSING:
            return self.default
        return self.default_factory()

    def __repr__(self) -> str:
        default = "MISSING" if self.get_default is MISSING else self.get_default
        return f"TableAttr(name={self.name}, type={self.type}, default={default})"

    def __lt__(self, other: Any):
        return Comparison(lambda e: getattr(e, self.name) < other)

    def __le__(self, other: Any):
        return Comparison(lambda e: getattr(e, self.name) <= other)

    def __eq__(self, other: Any):
        try:
            pattern = re.compile(other)
            return Comparison(
                lambda e: pattern.search(getattr(e, self.name)) is not None
            )
        except (re.error, TypeError):
            return Comparison(lambda e: getattr(e, self.name) == other)

    def __ne__(self, other: Any):
        try:
            pattern = re.compile(other)
            return Comparison(lambda e: pattern.search(getattr(e, self.name)) is None)
        except (re.error, TypeError):
            return Comparison(lambda e: getattr(e, self.name) != other)

    def __gt__(self, other: Any):
        return Comparison(lambda e: getattr(e, self.name) > other)

    def __ge__(self, other: Any):
        return Comparison(lambda e: getattr(e, self.name) >= other)


class MetaTable(type):
    def __getattribute__(self, name):
        if name != "__dataclass_fields__" and name in (
            dict := getattr(self, "__dataclass_fields__", {})
        ):
            return dict.setdefault(f"{name}_attr", TableField(dict[name]))
        return super().__getattribute__(name)


class Table(metaclass=MetaTable):
    @property
    def id(self):
        return self._id

    def __post_init__(self):
        self._id = -1

    def __eq__(self, other) -> bool:
        if self.__class__ == other.__class__:
            # Both objects are table entries
            if self.id == -1 and other.id == -1:
                # Both objects are not in database
                return super().__eq__(other)
            if self.id == other.id:
                # Both objects refer to same table entry
                return True
        return False


TableType = TypeVar("TableType", bound=Table)
ComparisonType = Union[Comparison, bool]


class Database:
    _file: Path = Path("data.json")
    _data: dict[str, Any] = None

    _encoders: dict[str, Encoder] = {}
    _decoders: dict[str, Decoder] = {}

    @classmethod
    def file(cls, new_file: Path = None) -> Path:
        if new_file is not None:
            cls._file = new_file
        return cls._file

    @classmethod
    def register_encoder(cls, encoder: Encoder) -> None:
        _type = encoder.obj_type.__name__
        if _type in cls._encoders:
            raise TypeError(f"Encoder for type '{_type}' already registered")
        cls._encoders[_type] = encoder

    @classmethod
    def register_decoder(cls, decoder: Decoder) -> None:
        _type = decoder.obj_type.__name__
        if _type in cls._decoders:
            raise TypeError(f"Decoder for type '{_type}' already registered")
        cls._decoders[_type] = decoder

    @classmethod
    def _encode(cls, obj: Any) -> dict[str, JsonValue]:
        found = [e for e in cls._encoders.values() if e.check(obj)]
        _len = len(found)
        if _len > 1:
            obj_type = obj.__class__.__name__
            TypeError(f"Multiple encoders found for {obj_type}: {found}")
        if _len == 1:
            return found[0].encode(obj)
        obj_type = obj.__class__.__name__
        raise TypeError(f"Object of type {obj_type} is not JSON serializable")

    @classmethod
    def _decode(cls, obj: dict[str, JsonValue]) -> Any:
        if "__type__" in obj:
            found = [d for d in cls._decoders.values() if d.check(obj)]
            _len = len(found)
            if _len > 1:
                obj_type = obj["__type__"]
                TypeError(f"Multiple decoders found for {obj_type}: {found}")
            if _len == 1:
                return found[0].decode(obj)
            obj_type = obj["__type__"]
            raise TypeError(f"Object of type {obj_type} is not JSON deserializable")
        return obj

    @classmethod
    def last_commit(cls) -> datetime:
        if cls._data is None:
            cls.rollback()
        return datetime.fromisoformat(cls._data["last_commit"])

    @classmethod
    def rollback(cls) -> None:
        try:
            cls._data = json.loads(cls._file.read_text(), object_hook=cls._decode)
        except (FileNotFoundError, json.decoder.JSONDecodeError):
            cls._data = {"last_commit": datetime.now()}

    @classmethod
    def commit(cls) -> bool:
        if cls._data is None:
            cls.rollback()

        previous_commit = (
            cls._data["last_commit"] if "last_commit" in cls._data else None
        )
        try:
            cls._data["last_commit"] = datetime.now()
            cls._file.write_text(json.dumps(cls._data, indent=2, default=cls._encode))
            return True
        except (IOError, Exception) as e:
            cls._data["last_commit"] = (
                datetime.now() if previous_commit is None else previous_commit
            )
            print(f"I/O error({e.errno}): {e.strerror}")
        return False

    @classmethod
    def get(
        cls, table: Type[TableType], where: ComparisonType = None
    ) -> tuple[TableType]:
        results: list[TableType] = []
        for id, entry in cls._get(table, where):
            entry._id = int(id)
            results.append(entry)
        return tuple(results)

    @classmethod
    def pop(
        cls, table: Type[TableType], where: ComparisonType = None
    ) -> tuple[TableType]:
        results: list[TableType] = []
        for id, table_entry in cls._get(table, where):
            results.append(table(**table_entry))
        return tuple(results)

    @classmethod
    def put(cls, table: Type[TableType], *entries: TableType):
        if cls._data is None:
            cls.rollback()

        table_name = table.__name__.lower()

        if table_name not in cls._data:
            cls._data[table_name] = {}

        table_data = cls._data[table_name]

        for entry in entries:
            if entry.id < 0:
                id = 0
                while str(id) in table_data:
                    id += 1
                entry._id = id
            table_data[str(entry.id)] = entry

    @classmethod
    def _get(
        cls, table: Type[TableType], where: ComparisonType = None
    ) -> list[tuple[str, TableType]]:
        if cls._data is None:
            cls.rollback()

        table_name = table.__name__.lower()
        if table_name not in cls._data:
            return []

        pre_check = where is None or (isinstance(where, bool) and where)

        results: list[tuple[str, TableType]] = []
        for id, table_entry in cls._data[table_name].items():
            if pre_check or where.compare(table_entry):
                results.append((id, table_entry))
        return results


for register, coder in {
    (
        Database.register_encoder,
        Encoder(datetime, lambda datetime: datetime.isoformat()),
    ),
    (
        Database.register_decoder,
        Decoder(datetime, lambda datetime_str: datetime.fromisoformat(datetime_str)),
    ),
    (Database.register_encoder, TableEncoder()),
    (Database.register_decoder, TableDecoder()),
}:
    try:
        register(coder)
    except KeyError:
        pass



from __future__ import annotations

import functools
import json
import logging
import pprint
from dataclasses import Field, asdict, dataclass, field, fields
from pathlib import Path
from typing import Optional

import warehub
from warehub.utils import parse_url

logger = logging.getLogger(warehub.__title__)


def _raise(field: Field, message: str) -> None:
    attrs = {a: getattr(field, a) for a in field.__slots__}
    raise ValueError(message.format(**attrs))


def _warn(field: Field, message: str) -> None:
    attrs = {a: getattr(field, a) for a in field.__slots__}
    logger.warning(message.format(**attrs))


@dataclass(frozen=True)
class Config:
    path: Path = field(
        default=".",
        metadata={
            "on_missing": functools.partial(
                _raise,
                message="Config must specify '{name}'. Default is '{default}'",
            ),
            "convert": (lambda string: Path(string).resolve()),
        },
    )
    database: str = field(
        default="data.json",
        metadata={
            "on_missing": functools.partial(
                _raise,
                message="Config must specify '{name}'. Default is '{default}'",
            ),
        },
    )
    url: str = field(
        default="",
        metadata={
            "on_missing": functools.partial(
                _raise,
                message="Config must specify '{name}'. Usually 'https://<user>.github.io/<repo_name/'",
            ),
            "convert": parse_url,
        },
    )
    title: Optional[str] = field(
        default="Personal Python Package Index",
        metadata={
            "on_missing": functools.partial(
                _warn,
                message="Config does not specify '{name}'. Using Default: '{default}'",
            ),
        },
    )
    description: Optional[str] = field(
        default="Welcome to your private Python package index!",
        metadata={
            "on_missing": functools.partial(
                _warn,
                message="Config does not specify '{name}'. Using Default: '{default}'",
            ),
        },
    )
    image_url: Optional[str] = field(
        default="https://pypi.org/static/images/logo-small.95de8436.svg",
        metadata={
            "on_missing": functools.partial(
                _warn,
                message="Config does not specify '{name}'. Using Default: '{default}'",
            ),
            "convert": parse_url,
        },
    )

    @classmethod
    def load(cls, file: Path):
        if not file.exists():
            logger.info("Generating Default Config")
            file.parent.mkdir(parents=True, exist_ok=True)
            file.write_text(json.dumps(asdict(cls()), indent=4))

        # This may error out if the json is malformed
        loaded: dict[str, str] = json.loads(file.read_text())

        for field in fields(cls):
            metadata = field.metadata
            if field.name not in loaded or loaded[field.name] == "":
                metadata["on_missing"](field)
                loaded[field.name] = field.default
            if "convert" in metadata:
                loaded[field.name] = metadata["convert"](loaded[field.name])

        if not loaded["url"].endswith("/"):
            loaded["url"] += "/"

        logger.info(pprint.pformat(cls(**loaded)))

        for key, value in loaded.items():
            setattr(cls, key, value)



from __future__ import annotations


class WarehubException(Exception):
    """Base Exception"""


class DatabaseException(WarehubException):
    """Base Database Exception"""


class InvalidDistribution(WarehubException):
    """Raised when a distribution is invalid."""



import cgi
import json
import logging
import pprint
import sys
import tempfile
from collections import defaultdict
from dataclasses import asdict
from pathlib import Path
from typing import DefaultDict, List, Optional

import readme_renderer.markdown
import readme_renderer.rst
import requests

import warehub
from warehub.arguments import AddArgs, Arguments, GenerateArgs, YankArgs
from warehub.config import Config
from warehub.database import Database
from warehub.model import Directory, File, Project, Release, Template
from warehub.package import Package, add_package
from warehub.utils import delete_path, file_size_str, parse_url

logger = logging.getLogger(warehub.__title__)


def init(args: List[str]):
    """Execute the ``init`` command.
    :param args:
        The command-line arguments.
    """

    generic_args: Arguments = Arguments.from_args(args)

    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(logging.DEBUG if generic_args.verbose else logging.WARNING)

    logger.info(pprint.pformat(generic_args))

    if not generic_args.config.exists():
        logger.info("Generating Default Config")
        generic_args.config.parent.mkdir(parents=True, exist_ok=True)
        generic_args.config.write_text(json.dumps(asdict(Config()), indent=4))

    logger.debug("Generating Database File")
    Database.file(Config.path / Config.database)
    Database.commit()

    logger.debug("Generating File Structure")
    for directory in Directory.LIST:
        (Config.path / directory).mkdir(parents=True, exist_ok=True)


# TODO - This should be renamed to github and add should be for specific file urls
def add(args: List[str]):
    """Execute the ``add`` command.
    :param args:
        The command-line arguments.
    """

    add_args: AddArgs = AddArgs.from_args(args)

    setup(add_args)

    return add_impl(add_args)


def generate(args: List[str]):
    """Execute the ``generate`` command.
    :param args:
        The command-line arguments.
    """

    generate_args: GenerateArgs = GenerateArgs.from_args(args)

    setup(generate_args)

    return generate_impl(generate_args)


def yank(args: List[str]):
    """Execute the ``yank`` command.
    :param args:
        The command-line arguments.
    """

    yank_args: YankArgs = YankArgs.from_args(args)

    setup(yank_args)

    return yank_impl(yank_args)


def setup(args: Arguments) -> None:
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(logging.DEBUG if args.verbose else logging.WARNING)

    logger.info(pprint.pformat(args))

    Config.load(args.config)

    Database.file(Config.path / Config.database)

    for directory in Directory.LIST:
        (Config.path / directory).mkdir(parents=True, exist_ok=True)


def add_impl(args: AddArgs):
    auth: Optional[tuple[Optional[str], Optional[str]]] = (
        (args.username or "", args.password or "")
        if args.username or args.password
        else None
    )
    if auth is not None:
        logger.debug("Authentication provided")

    file_urls: list[str] = []
    for repo_path in args.repositories:
        repo_url = parse_url(args.domain + f"repos/{repo_path}/releases")
        logger.info(f"Getting Releases from: {repo_url}")
        response = requests.get(repo_url, auth=auth)
        releases_obj = response.json()
        logger.debug(f"Response Code: {response.status_code}")
        if response.status_code != requests.codes.ok:
            logger.warning(
                f"Could not get information on release for "
                f"'{repo_url}': {releases_obj['message']}"
            )
            continue
        for info in releases_obj:
            for asset in info["assets"]:
                file_urls.append(asset["browser_download_url"])
    if len(file_urls) == 0:
        logger.info("No Files to Download")
        return

    logger.info(f"Found {len(file_urls)} files")
    with tempfile.TemporaryDirectory() as temp:
        temp_dir = Path(temp)

        downloaded_files: list[Path] = []
        for file_url in file_urls:
            download = requests.get(file_url, auth=auth)
            logger.debug(f"Response Code: {download.status_code}")
            if download.status_code != requests.codes.ok:
                logger.warning(
                    f"Could not download '{file_url}': " f"{download.status_code}"
                )
                continue

            file = temp_dir / file_url.split("/")[-1]
            file.write_bytes(download.content)

            logger.info(f"Downloaded File: {file_url}\n  to: {file.absolute()}")

            downloaded_files.append(file)

        signatures: dict[str, Path] = {
            f.name: f for f in downloaded_files if f.suffix == ".asc"
        }
        logger.debug(f"Signature Files:\n" + pprint.pformat(signatures))

        added: set[str] = set()
        for file in downloaded_files:
            try:
                if file.suffix != ".asc":
                    package = Package(file, None)

                    if (signed_name := package.signed_file.name) in signatures:
                        package.gpg_signature = signatures[signed_name]

                    logger.debug(
                        f"Package created for file: '{package.file.name}' "
                        f"({file_size_str(package.file)})"
                    )
                    if package.gpg_signature:
                        logger.debug(f"\tSigned with {package.signed_file}")

                    add_package(package)

                    added.add(Config.url + f"simple/{package.name}/{package.version}/")

            except Exception as e:
                logger.exception(
                    f"Exception found when processing file: {file.name}", exc_info=e
                )

        if len(added) > 0:
            if not args.no_generate:
                generate_args: GenerateArgs = GenerateArgs(args.verbose, args.config)

                logger.info(pprint.pformat(generate_args))

                generate_impl(generate_args)

            logger.info(f"View new Packages at:")
            for url in added:
                logger.info(f"\t{url}")


def generate_impl(args: GenerateArgs):
    logger.debug("Deleting Existing Files")
    for directory in [Directory.PROJECT, Directory.SIMPLE, Directory.PYPI]:
        for child in (Config.path / directory).glob("*"):
            delete_path(child)

    logger.info("Generating Homepage")
    projects_listing = ""
    for project in Database.get(Project):
        releases = Database.get(Release, where=Release.project_id == project.id)

        if len(releases) < 1:
            # This should never happen because a project is always created
            # with at least one release, but you never know...
            logger.warning(f"Projects does not have any release: {project}")
            continue

        latest = releases[0]
        for release in releases:
            if not release.yanked and release.version > latest.version:
                latest = release

        indent = " " * 4
        projects_listing += (
            f'\n{indent}<a class="card" href="{Config.url}{Directory.PROJECT}/{project.name}/">'
            f'\n{indent}    {project.name}<span class="version">{latest.version}</span>'
            f'\n{indent}    <span class="description">{latest.summary}</span>'
            f"\n{indent}</a>"
        )

    homepage_template = Template.HOMEPAGE

    for string, value in [
        ("%%WAREHUB_VERSION%%", warehub.__version__),
        ("%%STYLE%%", "".join("\n        " + s for s in Template.STYLE.splitlines())),
        ("%%URL%%", Config.url),
        ("%%TITLE%%", Config.title),
        ("%%DESCRIPTION%%", Config.description),
        ("%%IMAGE%%", Config.image_url),
        ("%%PACKAGES%%", projects_listing),
    ]:
        homepage_template = homepage_template.replace(string, value)

    homepage = Config.path / "index.html"
    homepage.write_text(homepage_template)

    logger.info("Generating Project Release Pages")

    def generate_release_page(
        path: Path, project: Project, release: Release, show_version: bool = False
    ):
        renderers = {
            "text/plain": None,
            "text/x-rst": readme_renderer.rst,
            "text/markdown": readme_renderer.markdown,
        }

        links = ""
        for name, url in release.urls.items():
            indent = " " * 20
            links += f'\n{indent}<li><a href="{url}" rel="nofollow">{name}</a></li>'

        meta = ""
        for name, check, string in {
            ("License", release.license, release.license),
            (
                "Author",
                release.author,
                f'<a href="mailto:{release.author_email}">{release.author}</a>',
            ),
            (
                "Maintainer",
                release.maintainer,
                f'<a href="mailto:{release.maintainer_email}">{release.maintainer}</a>',
            ),
            ("Requires", release.requires_python, release.requires_python),
            ("Platform", release.platform, release.platform),
        }:
            if check is not None:
                indent = " " * 16
                meta += (
                    f'\n{indent}<p class="elem"><strong>{name}: </strong>{string}</p>'
                )

        classifiers: DefaultDict[str, list[str]] = defaultdict(list)
        for classifier in release.classifiers:
            group, tag = classifier.split(" :: ", 1)
            classifiers[group].append(tag)

        classifiers_str = ""
        for group, tags in classifiers.items():
            tags_str = ""
            for tag in sorted(tags):
                indent = " " * 28
                tags_str += f"\n{indent}<li>{tag}</li>"
            indent = " " * 20
            classifiers_str += "\n".join(
                [
                    f"",
                    f"{indent}<li>",
                    f"{indent}    <strong>{group}</strong>",
                    f"{indent}    <ul>{tags_str}",
                    f"{indent}    </ul>",
                    f"{indent}</li>",
                ]
            )

        description = release.description["raw"]
        content_type, params = cgi.parse_header(release.description["content_type"])
        renderer = renderers.get(content_type, readme_renderer.rst)

        if description in {None, "UNKNOWN\n\n\n"}:
            description = ""
        elif renderer:
            description = renderer.render(description, **params) or ""

        releases = ""
        for r in Database.get(Release, where=Release.project_id == project.id):
            indent = " " * 16
            # TODO - Pre-Release?, Yanked?, etc
            releases += "\n".join(
                [
                    f"",
                    f'{indent}<a class="card" href="{Config.url}{Directory.PROJECT}/{project.name}/{r.version}/">',
                    f'{indent}    <span class="version">{r.version}</span>',
                    f"{indent}</a>",
                ]
            )

        files = ""
        for f in Database.get(File, where=File.release_id == release.id):
            indent = " " * 16
            files += "\n".join(
                [
                    f"",
                    f'{indent}<a class="card" href="{Config.url}{Directory.FILES}/{f.name}">',
                    f"{indent}    {f.name}",
                    f"{indent}</a>",
                ]
            )

        template = Template.RELEASE

        for string, value in [
            ("%%WAREHUB_VERSION%%", warehub.__version__),
            (
                "%%STYLE%%",
                "".join("\n        " + s for s in Template.STYLE.splitlines()),
            ),
            ("%%URL%%", Config.url),
            ("%%TITLE%%", Config.title),
            ("%%IMAGE%%", Config.image_url),
            ("%%NAME%%", project.name),
            ("%%VERSION%%", release.version),
            ("%%PIP_VERSION%%", f"=={release.version}" if show_version else ""),
            ("%%SUMMARY%%", release.summary or ""),
            ("%%LINKS%%", links),
            ("%%META%%", meta),
            ("%%CLASSIFIERS%%", classifiers_str),
            ("%%DESCRIPTION%%", description),
            ("%%RELEASES%%", releases),
            ("%%FILES%%", files),
        ]:
            template = template.replace(string, value)

        file = path / "index.html"
        file.write_text(template)

    for project in Database.get(Project):
        releases = Database.get(Release, where=Release.project_id == project.id)

        if len(releases) < 1:
            # This should never happen because a project is always created
            # with at least one release, but you never know...
            logger.warning("Projects does not have any release:", project)

        project_dir = Config.path / Directory.PROJECT / project.name
        project_dir.mkdir(parents=True, exist_ok=True)

        latest = releases[0]
        for release in releases:
            if release.version > latest.version:
                latest = release

            release_dir = project_dir / release.version
            release_dir.mkdir(parents=True, exist_ok=True)

            logger.info(f"Generating Release Page for: {release}")
            generate_release_page(release_dir, project, release, True)

        logger.info(f"Generating Project Page for: {project}")
        generate_release_page(project_dir, project, latest)

    logger.info("Generating Simple Pages")

    def create_simple(link_list: str) -> str:
        template = Template.SIMPLE

        for string, value in [
            ("%%WAREHUB_VERSION%%", warehub.__version__),
            ("%%TITLE%%", Config.title),
            ("%%IMAGE%%", Config.image_url),
            ("%%LIST%%", link_list),
        ]:
            template = template.replace(string, value)
        return template

    project_list = ""
    for project in Database.get(Project):
        releases = Database.get(Release, where=Release.project_id == project.id)

        if len(releases) < 1:
            # This should never happen because a project is always created
            # with at least one release, but you never know...
            logger.warning("Projects does not have any release:", project)
            continue

        f_list = ""
        for release in releases:
            if not release.yanked:
                for file in Database.get(File, where=File.release_id == release.id):
                    f_list += f'\n    <a href="{Config.url}{Directory.FILES}/{file.name}">{file.name}</a><br/>'

        project_dir = Config.path / Directory.SIMPLE / project.name
        project_dir.mkdir(parents=True, exist_ok=True)

        logger.info(f"Generating Simple File list for {project}")
        template = create_simple(f_list)

        project_file = project_dir / "index.html"
        project_file.write_text(template)

        project_list += (
            f'\n    <a class="card" href="{project.name}/">{project.name}</a><br/>'
        )

    logger.info("Generating Simple Project List")
    template = create_simple(project_list)

    landing = Config.path / Directory.SIMPLE / "index.html"
    landing.write_text(template)

    logger.info("Generating Json Pages")

    def create_json(path: Path, project: Project, release: Release):
        dir = path / "json"
        dir.mkdir(parents=True, exist_ok=True)

        releases = {}
        for r in Database.get(Release, where=Release.project_id == project.id):
            releases[r.version] = [
                {
                    "filename": f.name,
                    "python_version": f.python_version,
                    "packagetype": f.package_type,
                    "comment_text": f.comment_text,
                    "size": f.size,
                    "has_sig": f.has_signature,
                    "md5_digest": f.md5_digest,
                    "digests": {
                        "md5": f.md5_digest,
                        "sha256": f.sha256_digest,
                        "blake2_256": f.blake2_256_digest,
                    },
                    "downloads": (-1),
                    "upload_time": f.upload_time.strftime("%Y-%m-%dT%H:%M:%S"),
                    "upload_time_iso_8601": f.upload_time.strftime(
                        "%Y-%m-%dT%H:%M:%S.%fZ"
                    ),
                    "url": f"{Config.url}{Directory.FILES}/{f.name}",
                    "requires_python": r.requires_python if r.requires_python else None,
                    "yanked": r.yanked,
                    "yanked_reason": r.yanked_reason or None,
                }
                for f in Database.get(File, where=File.release_id == r.id)
            ]

        file = dir / "index.json"
        file.write_text(
            json.dumps(
                {
                    "info": {
                        "name": project.name,
                        "version": release.version,
                        "summary": release.summary,
                        "description_content_type": release.description["content_type"],
                        "description": release.description["raw"],
                        "keywords": release.keywords,
                        "license": release.license,
                        "classifiers": release.classifiers,
                        "author": release.author,
                        "author_email": release.author_email,
                        "maintainer": release.maintainer,
                        "maintainer_email": release.maintainer_email,
                        "requires_python": release.requires_python,
                        "platform": release.platform,
                        "downloads": {
                            "last_day": -1,
                            "last_week": -1,
                            "last_month": -1,
                        },
                        "package_url": f"{Config.url}{Directory.PROJECT}/{project.name}",
                        "project_url": f"{Config.url}{Directory.PROJECT}/{project.name}",
                        "project_urls": release.urls,
                        "release_url": f"{Config.url}{Directory.PROJECT}/{project.name}/{release.version}",
                        "requires_dist": release.dependencies["requires_dist"],
                        "docs_url": None,
                        "bugtrack_url": None,
                        "home_page": release.home_page,
                        "download_url": release.download_url,
                        "yanked": release.yanked,
                        "yanked_reason": release.yanked_reason or None,
                    },
                    "urls": releases[release.version],
                    "releases": releases,
                    "vulnerabilities": [],
                    "last_serial": (-1),
                },
                indent=4,
            )
        )

    for project in Database.get(Project):
        releases = Database.get(Release, where=Release.project_id == project.id)

        if len(releases) < 1:
            # This should never happen because a project is always created
            # with at least one release, but you never know...
            logger.warning("Projects does not have any release:", project)
            continue

        project_dir = Config.path / Directory.PYPI / project.name
        project_dir.mkdir(parents=True, exist_ok=True)

        latest: Optional[Release] = None
        for release in releases:
            if not release.yanked:
                if latest is None or release.version > latest.version:
                    latest = release

                create_json(project_dir / release.version, project, release)

        if latest is not None:
            create_json(project_dir, project, latest)


def yank_impl(args: YankArgs):
    pass



from __future__ import annotations

import email.utils
import hashlib
import io
import re
import shutil
from cgi import parse_header
from pathlib import Path
from types import NoneType
from typing import Any, Optional, Type, Union, get_args, get_origin

from packaging.requirements import InvalidRequirement, Requirement
from packaging.specifiers import InvalidSpecifier, SpecifierSet
from packaging.version import Version
from pkginfo import BDist, SDist, Wheel
from pkginfo.distribution import HEADER_ATTRS
from rfc3986 import uri_reference
from rfc3986.exceptions import ValidationError
from rfc3986.validators import Validator
from trove_classifiers import classifiers, deprecated_classifiers

from warehub.config import Config
from warehub.database import Database
from warehub.exceptions import InvalidDistribution
from warehub.model import Directory, File, FileName, Project, Release
from warehub.utils import file_size_str

ONE_KB = 1024
ONE_MB = ONE_KB * ONE_KB
ONE_GB = ONE_KB * ONE_KB * ONE_KB

MAX_SIG_SIZE = 8 * ONE_KB
MAX_FILE_SIZE = 100 * ONE_MB
MAX_PROJECT_SIZE = 10 * ONE_GB

DIST_TYPES = {
    "bdist_wheel": Wheel,
    "bdist_egg": BDist,
    "sdist": SDist,
}

DIST_EXTENSIONS = {
    ".whl": "bdist_wheel",
    ".egg": "bdist_egg",
    ".tar.bz2": "sdist",
    ".tar.gz": "sdist",
    ".zip": "sdist",
}

DIST_VERSION = {
    "bdist_wheel": re.compile(
        (
            r"^(?P<namever>(?P<name>.+?)(-(?P<ver>\d.+?))?)"
            r"((-(?P<build>\d.*?))?-(?P<pyver>.+?)-(?P<abi>.+?)-(?P<plat>.+?)"
            r"\.whl|\.dist-info)$"
        ),
        re.VERBOSE,
    ),
    "bdist_egg": re.compile(
        (
            r"^(?P<namever>(?P<name>.+?)(-(?P<ver>\d.+?))?)"
            r"((-(?P<build>\d.*?))?-(?P<pyver>.+?)-(?P<abi>.+?)-(?P<plat>.+?)"
            r"\.egg|\.egg-info)$"
        ),
        re.VERBOSE,
    ),
}


def add_package(package: Package) -> None:
    projects = Database.get(Project, where=Project.name == package.name)
    if len(projects) == 0:
        project = Project(
            name=package.name,
        )
        Database.put(Project, project)
    elif len(projects) > 1:
        raise ValueError(f"Multiple Projects found with name '{package.name}'")
    else:
        project = projects[0]

    releases = Database.get(
        Release,
        where=(
            (Release.project_id == project.id) & (Release.version == package.version)
        ),
    )
    if len(releases) == 0:
        release = Release(
            project_id=project.id,
            version=package.version,
            author=package.author,
            author_email=package.author_email,
            maintainer=package.maintainer,
            maintainer_email=package.maintainer_email,
            summary=package.summary,
            description={
                "raw": package.description or "",
                "content_type": package.description_content_type,
            },
            keywords=package.keywords,
            classifiers=package.classifiers,
            license=package.license,
            platform=package.platform,
            home_page=package.home_page,
            download_url=package.download_url,
            requires_python=package.requires_python,
            dependencies={
                "requires": package.requires or [],
                "provides": package.provides or [],
                "obsoletes": package.obsoletes or [],
                "requires_dist": package.requires_dist or [],
                "provides_dist": package.provides_dist or [],
                "obsoletes_dist": package.obsoletes_dist or [],
                "requires_external": package.requires_external or [],
            },
            project_urls=package.project_urls or [],
            # uploader=package.uploader,
            # uploaded_via=package.uploaded_via,
            # yanked=package.yanked,
            # yanked_reason=package.yanked_reason,
        )
        Database.put(Release, release)
    elif len(releases) > 1:
        raise ValueError(f"Multiple Releases found with name '{package.name}'")
    else:
        release = releases[0]

    filenames = Database.get(FileName, where=FileName.name == package.file.name)
    if len(filenames) > 0:
        raise ValueError(f"File already exists with that name: {package.file.name}")
    Database.put(FileName, FileName(package.file.name))

    # TODO - Check for multiple sdist
    # TODO - Check for valid dist file
    # TODO - Check that if it's a binary wheel, it's on a supported platform

    if package.gpg_signature is not None:
        has_signature = True

        filenames = Database.get(
            FileName, where=FileName.name == package.signed_file.name
        )
        if len(filenames) > 0:
            raise ValueError(
                f"File already exists with that name: {package.signed_file.name}"
            )

        Database.put(FileName, FileName(package.signed_file.name))
    else:
        has_signature = False

    file = File(
        release_id=release.id,
        name=package.file.name,
        python_version=package.pyversion,
        package_type=package.filetype,
        comment_text=package.comment,
        size=package.file.stat().st_size,
        has_signature=has_signature,
        md5_digest=package.md5_digest,
        sha256_digest=package.sha256_digest,
        blake2_256_digest=package.blake2_256_digest,
        # uploaded_via=,
    )
    Database.put(File, file)

    file_size = package.file.stat().st_size
    if file_size > MAX_FILE_SIZE:
        raise ValueError(f"File too large. Limit is {file_size_str(MAX_FILE_SIZE)}")
    project.total_size += file_size

    if has_signature:
        file_size = package.signed_file.stat().st_size
        if file_size > MAX_SIG_SIZE:
            raise ValueError(
                f"Signature file too large. Limit is {file_size_str(MAX_SIG_SIZE)}"
            )
        project.total_size += file_size

    if project.total_size > MAX_PROJECT_SIZE:
        raise ValueError(
            f"Project is now too large. Limit is {file_size_str(MAX_PROJECT_SIZE)}"
        )

    shutil.copy(package.file, Config.path / Directory.FILES / package.file.name)
    if has_signature:
        shutil.copy(
            package.signed_file,
            Config.path / Directory.FILES / package.signed_file.name,
        )
    Database.commit()


class Package:
    def __init__(self, file: Path, comment: Optional[str]):
        self.file: Path = file
        self.comment: Optional[str] = comment

        self.signed_file: Path = self.file.with_name(self.file.name + ".asc")
        self.gpg_signature: Optional[Path] = None

        hashes = {
            "md5": hashlib.md5(),
            "sha256": hashlib.sha256(),
            "blake2_256": hashlib.blake2b(digest_size=256 // 8),
        }

        with self.file.open("rb") as fp:
            for content in iter(lambda: fp.read(io.DEFAULT_BUFFER_SIZE), b""):
                for hash in hashes.values():
                    hash.update(content)

        self.md5_digest: str = hashes["md5"].hexdigest().lower()
        self.sha256_digest: str = validate_hash(
            "Use a valid, hex-encoded, SHA256 message digest.",
            hashes["sha256"].hexdigest().lower(),
        )
        self.blake2_256_digest: str = validate_hash(
            "Use a valid, hex-encoded, BLAKE2 message digest.",
            hashes["blake2_256"].hexdigest().lower(),
        )

        self.filetype: str = ""
        for ext, file_type in DIST_EXTENSIONS.items():
            if self.file.name.endswith(ext):
                try:
                    metadata = DIST_TYPES[file_type](self.file)  # Convert to str?
                except EOFError:
                    raise InvalidDistribution(
                        f"Invalid distribution file: '{self.file.name}'"
                    )
                else:
                    self.filetype = file_type
                    break
        else:
            raise InvalidDistribution(
                f"Unknown distribution format: '{self.file.name}'"
            )

        # If pkginfo encounters a metadata version it doesn't support, it may
        # give us back empty metadata. At the very least, we should have a name
        # and version
        if not (metadata.name and metadata.version):
            supported_metadata = list(HEADER_ATTRS)
            raise InvalidDistribution(
                "Invalid distribution metadata. "
                "This version of twine supports Metadata-Version "
                f'{", ".join(supported_metadata[:-1])}, and {supported_metadata[-1]}'
            )

        self.pyversion: Optional[str] = None
        if self.filetype in DIST_VERSION:
            self.pyversion = "any"
            if (m := DIST_VERSION[self.filetype].match(self.file.name)) is not None:
                self.pyversion = m.group("pyver")

        self.metadata_version: str = validate_metadata_version(
            sanitize(str, metadata.metadata_version)
        )
        # version 1.0
        self.name: str = validate_name(
            re.sub("[^A-Za-z0-9.]+", "-", sanitize(str, metadata.name))
        )
        self.version: str = str(Version(sanitize(str, metadata.version)))
        self.author: Optional[str] = sanitize(Optional[str], metadata.author)
        self.author_email: Optional[str] = validate_email(
            sanitize(Optional[str], metadata.author_email)
        )
        self.summary: Optional[str] = validate_summary(
            sanitize(Optional[str], metadata.summary)
        )
        self.description: Optional[str] = sanitize(Optional[str], metadata.description)
        self.keywords: Optional[str] = sanitize(Optional[str], metadata.keywords)
        self.license: Optional[str] = sanitize(Optional[str], metadata.license)
        self.platform: Optional[str] = sanitize(Optional[str], metadata.platforms)
        self.home_page: Optional[str] = validate_uri(
            sanitize(Optional[str], metadata.home_page)
        )
        self.download_url: Optional[str] = validate_uri(
            sanitize(Optional[str], metadata.download_url)
        )
        self.supported_platforms: Optional[str] = sanitize(
            Optional[str], metadata.supported_platforms
        )
        # version 1.1
        self.classifiers: list = validate_classifiers(
            sanitize(list, metadata.classifiers)
        )
        self.requires: Optional[list] = validate_legacy_non_dist(
            sanitize(Optional[list], metadata.requires)
        )
        self.provides: Optional[list] = validate_legacy_non_dist(
            sanitize(Optional[list], metadata.provides)
        )
        self.obsoletes: Optional[list] = validate_legacy_non_dist(
            sanitize(Optional[list], metadata.obsoletes)
        )
        # version 1.2
        self.maintainer: Optional[str] = sanitize(Optional[str], metadata.maintainer)
        self.maintainer_email: Optional[str] = validate_email(
            sanitize(Optional[str], metadata.maintainer_email)
        )
        self.requires_python: Optional[str] = validate_pep440_specifier(
            sanitize(Optional[str], metadata.requires_python)
        )
        self.requires_dist: Optional[list] = validate_legacy_dist(
            sanitize(Optional[list], metadata.requires_dist)
        )
        self.provides_dist: Optional[list] = validate_legacy_dist(
            sanitize(Optional[list], metadata.provides_dist)
        )
        self.obsoletes_dist: Optional[list] = validate_legacy_dist(
            sanitize(Optional[list], metadata.obsoletes_dist)
        )
        self.requires_external: Optional[list] = validate_requires_external(
            sanitize(Optional[list], metadata.requires_external)
        )
        self.project_urls: Optional[list] = validate_project_urls(
            sanitize(Optional[list], metadata.project_urls)
        )
        # version 2.1
        self.description_content_type: Optional[
            str
        ] = validate_description_content_type(
            sanitize(Optional[str], metadata.description_content_type)
        )
        self.provides_extras: Optional[list] = sanitize(
            Optional[list], metadata.provides_extras
        )
        # version 2.2
        self.dynamic: Optional[list] = sanitize(Optional[list], metadata.dynamic)

    def add_gpg_signature(self, signature_file: Path) -> None:
        if self.gpg_signature is not None:
            raise InvalidDistribution("GPG Signature can only be added once")

        self.gpg_signature = signature_file


def validate_hash(message: str, hash: str) -> str:
    if re.match(r"^[A-F0-9]{64}$", hash, re.IGNORECASE) is None:
        raise ValueError(message)
    return hash


def sanitize(type: Type, value: Any) -> Any:
    types = get_args(type)
    optional = get_origin(type) is Union and NoneType in types
    type = types[0] if optional else type

    if isinstance(value, (list, tuple)) and type not in (list, tuple):
        value = value[0] if len(value) > 0 else None

    if value is None:
        if not optional:
            raise ValueError(f"missing required field")
    elif not isinstance(value, type):
        value = type(value)

    if isinstance(value, str):
        if value.strip() == "UNKNOWN":
            value = None
        elif "\x00" in value:
            value = value.replace("\x00", "\\x00")
    return value


def validate_metadata_version(value):
    if value not in ["1.0", "1.1", "1.2", "2.0", "2.1"]:
        raise ValueError("Use a known metadata version.")
    return value


def validate_name(value):
    if (
        re.match(r"^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$", value, re.IGNORECASE)
        is None
    ):
        raise ValueError(
            "Start and end with a letter or numeral containing "
            "only ASCII numeric and '.', '_' and '-'."
        )
    return value


def validate_email(value):
    if value is None:
        return value
    pattern = re.compile(
        (
            r"([a-z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*|\""
            r"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21\x23-\x5b\x5d-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])*\")"
            r"@((?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?"
            r"|\[(?:(?:2(?:5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9])\.){3}"
            r"(?:(?:2(?:5[0-5]|[0-4][0-9])|1[0-9][0-9]|[1-9]?[0-9])|[a-z0-9-]*[a-z0-9]:"
            r"(?:[\x01-\x08\x0b\x0c\x0e-\x1f\x21-\x5a\x53-\x7f]|\\[\x01-\x09\x0b\x0c\x0e-\x7f])+)])"
        ),
        re.IGNORECASE,
    )

    addresses = email.utils.getaddresses([value])

    for real_name, address in addresses:
        if pattern.match(address) is None:
            raise ValueError("Use a valid email address")
    return value


def validate_summary(value):
    if value is None:
        return value
    if len(value or "") > 512:
        raise ValueError("Summary is too long")
    if re.match(r"^.+$", value) is None:
        raise ValueError("Use a single line only.")
    return value


def validate_description_content_type(value):
    def _raise(message):
        raise ValueError(f"Invalid description content type: {message}")

    content_type, parameters = parse_header(value)
    if content_type not in {"text/plain", "text/x-rst", "text/markdown"}:
        _raise("type/subtype is not valid")

    charset = parameters.get("charset")
    if charset and charset != "UTF-8":
        _raise("Use a valid charset")

    valid_markdown_variants = {"CommonMark", "GFM"}

    variant = parameters.get("variant")
    if (
        content_type == "text/markdown"
        and variant
        and variant not in valid_markdown_variants
    ):
        _raise(
            f'Use a valid variant, expected one of {", ".join(valid_markdown_variants)}'
        )
    return value


def validate_classifiers(value):
    invalid_classifiers = set(value or []) & deprecated_classifiers.keys()
    if invalid_classifiers:
        first_invalid_classifier_name = sorted(invalid_classifiers)[0]
        deprecated_by = deprecated_classifiers[first_invalid_classifier_name]

        if deprecated_by:
            raise ValueError(
                f"Classifier {first_invalid_classifier_name!r} has been "
                "deprecated, use the following classifier(s) instead: "
                f"{deprecated_by}"
            )
        else:
            raise ValueError(
                f"Classifier {first_invalid_classifier_name!r} has been deprecated."
            )

    invalid = sorted(set(value or []) - classifiers)

    if invalid:
        if len(invalid) == 1:
            raise ValueError(f"Classifier {invalid[0]!r} is not a valid classifier.")
        else:
            raise ValueError(f"Classifiers {invalid!r} are not valid classifiers.")
    return value


def validate_uri(value):
    if value is None:
        return value
    if not is_valid_uri(value):
        raise ValueError(f"Invalid URI '{value}'")
    return value


def validate_pep440_specifier(value):
    try:
        SpecifierSet(value)
    except InvalidSpecifier:
        raise ValueError("Invalid specifier in requirement.") from None


def validate_legacy_non_dist(value):
    for datum in value:
        try:
            req = Requirement(datum.replace("_", ""))
        except InvalidRequirement:
            raise ValueError("Invalid requirement: {!r}".format(datum)) from None

        if req.url is not None:
            raise ValueError("Can't direct dependency: {!r}".format(datum))

        if any(
            not identifier.isalnum() or identifier[0].isdigit()
            for identifier in req.name.split(".")
        ):
            raise ValueError("Use a valid Python identifier.")
    return value


def validate_legacy_dist(value):
    for datum in value:
        try:
            req = Requirement(datum)
        except InvalidRequirement:
            raise ValueError("Invalid requirement: {!r}.".format(datum)) from None

        if req.url is not None:
            raise ValueError("Can't have direct dependency: {!r}".format(datum))
    return value


def validate_requires_external(value):
    for datum in value:
        parsed = re.search(r"^(?P<name>\S+)(?: \((?P<specifier>\S+)\))?$", datum)
        if parsed is None:
            raise ValueError("Invalid requirement.")
        name, specifier = parsed.groupdict()["name"], parsed.groupdict()["specifier"]

        if specifier is not None:
            validate_pep440_specifier(specifier)
    return value


def validate_project_urls(value):
    for datum in value:
        try:
            label, url = datum.split(", ", 1)
        except ValueError:
            raise ValueError("Use both a label and an URL.") from None

        if not label:
            raise ValueError("Use a label.")

        if len(label) > 32:
            raise ValueError("Use 32 characters or less.")

        if not url:
            raise ValueError("Use an URL.")

        if not is_valid_uri(str(url), require_authority=False):
            raise ValueError("Use valid URL.")
    return value


def is_valid_uri(
    uri,
    require_scheme: bool = True,
    allowed_schemes: list[str] = None,
    require_authority: bool = True,
):
    if allowed_schemes is None:
        allowed_schemes = ["http", "https"]

    uri = uri_reference(uri).normalize()
    validator = Validator().allow_schemes(*allowed_schemes)
    if require_scheme:
        validator.require_presence_of("scheme")
    if require_authority:
        validator.require_presence_of("host")

    validator.check_validity_of("scheme", "host", "port", "path", "query")

    try:
        validator.validate(uri)
    except ValidationError:
        return False

    return True



import importlib.metadata

__all__ = (
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
)

metadata = importlib.metadata.metadata("warehub")

__title__ = metadata["name"]
__summary__ = metadata["summary"]
__uri__ = metadata["home_page"]
__version__ = metadata["version"]
__author__ = metadata["author"]
__email__ = metadata["author_email"]
__license__ = metadata["license"]

__copyright__ = "Copyright 2022 Ryan Smith"



import argparse
import http
import importlib.metadata
import sys
from typing import Any

import colorama
import requests
from packaging.requirements import Requirement

import warehub
from warehub.exceptions import WarehubException


def list_deps_and_versions() -> list[tuple[str, str]]:
    requires = importlib.metadata.requires("warehub")
    deps = [Requirement(r).name for r in requires]
    return [(dep, importlib.metadata.version(dep)) for dep in deps]


def dep_versions() -> str:
    return ", ".join(
        "{}: {}".format(*dependency) for dependency in list_deps_and_versions()
    )


def main() -> Any:
    entry_points = importlib.metadata.entry_points()
    # noinspection PyUnresolvedReferences
    commands = entry_points.select(group="warehub.commands")

    parser = argparse.ArgumentParser(prog="warehub")

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s version {warehub.__version__} ({dep_versions()})",
    )
    parser.add_argument(
        "--no-color",
        default=False,
        required=False,
        action="store_true",
        help="disable colored output",
    )
    parser.add_argument(
        "command",
        choices=commands.names,
    )
    parser.add_argument(
        "args",
        help=argparse.SUPPRESS,
        nargs=argparse.REMAINDER,
    )

    args = parser.parse_args(sys.argv[1:])

    try:
        main = commands[args.command].load()

        result = main(args.args)
    except requests.HTTPError as exc:
        status_code = exc.response.status_code
        status_phrase = http.HTTPStatus(status_code).phrase
        result = (
            f"{exc.__class__.__name__}: {status_code} {status_phrase}"
            f" from {exc.response.url}\n"
            f"{exc.response.reason}"
        )
    except WarehubException as exc:
        result = f"{exc.__class__.__name__}: {exc.args[0]}"

    if isinstance(result, str):
        pre_style, post_style = "", ""
        if not args.no_color:
            colorama.init()
            pre_style, post_style = colorama.Fore.RED, colorama.Style.RESET_ALL
        return f"{pre_style}{result}{post_style}"
    return result


if __name__ == "__main__":
    sys.exit(main())



from pathlib import Path


def main():
    test_data_dir: Path = Path("test_data")

    data_file = test_data_dir / "data.json"
    data_file.unlink(missing_ok=True)


if __name__ == "__main__":
    main()



from pathlib import Path

from warehub.database import Database
from warehub.model import Project


def main():
    Database.file(Path("test/data.json"))

    # project = Project('Test')
    # Database.put(Project, project)

    projects = Database.get(Project, where=Project.name == "Test1")
    if len(projects) > 0:
        projects[0].name = "RENAMED"

    # projects = Database.get(Project)
    # pprint(projects)
    #
    # Database.commit()


if __name__ == "__main__":
    main()



from __future__ import annotations

from argparse import ArgumentParser
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Any, List, Optional, Type, TypeVar

import warehub
from warehub.utils import EnvironmentDefault, parse_url

T = TypeVar("T")


class Secure(str):
    def __repr__(self) -> str:
        return "***"


@dataclass(frozen=True)
class Arguments:
    verbose: bool = field(
        metadata={
            "name_or_flags": ["-v", "--verbose"],
            "default": False,
            "required": False,
            "action": "store_true",
            "help": "show verbose output",
        }
    )
    config: Path = field(
        metadata={
            "name_or_flags": ["-c", "--config"],
            "default": "./config.json",
            "required": False,
            "help": "The path to the config file. [default: ./config.json]",
            "convert": (lambda string: Path(string).resolve()),
        }
    )

    @classmethod
    def from_args(cls: Type[T], args: list[str]) -> T:
        """Generate the Settings from parsed arguments."""

        parser = ArgumentParser(prog=f"{warehub.__title__} {cls.__name__.lower()}")

        for field in fields(cls):
            metadata = dict(field.metadata)
            metadata.pop("convert", None)
            name_or_flags = tuple(
                metadata.pop("name_or_flags") if "name_or_flags" in metadata else []
            )
            parser.add_argument(*name_or_flags, **metadata)

        parsed = vars(parser.parse_args(args))

        for field in fields(cls):
            metadata = field.metadata
            if "convert" in metadata and (value := parsed[field.name]) is not None:
                parsed[field.name] = metadata["convert"](value)

        return cls._format(parsed)

    @classmethod
    def _format(cls: Type[T], parsed: dict[str, Any]) -> T:
        return cls(**parsed)


@dataclass(frozen=True)
class AddArgs(Arguments):
    username: Optional[str] = field(
        metadata={
            "name_or_flags": ["-u", "--username"],
            "action": EnvironmentDefault,
            "env": "WAREHUB_USERNAME",
            "required": False,
            "help": "The username to authenticate to the repository "
            "(package index) as. (Can also be set via "
            "%(env)s environment variable.) "
            "[default: env.WAREHUB_USERNAME]",
            "convert": Secure,
        }
    )
    password: Optional[str] = field(
        metadata={
            "name_or_flags": ["-p", "--password"],
            "action": EnvironmentDefault,
            "env": "WAREHUB_PASSWORD",
            "required": False,
            "help": "The password to authenticate to the repository "
            "(package index) with. (Can also be set via "
            "%(env)s environment variable.) "
            "[default: env.WAREHUB_PASSWORD]",
            "convert": Secure,
        }
    )
    domain: str = field(
        metadata={
            "name_or_flags": ["-d", "--domain"],
            "default": "https://api.github.com/",
            "required": False,
            "help": "The domain to access the Github api from. This "
            "will only change for Github Enterprise users. "
            "[default: https://api.github.com/]",
            "convert": parse_url,
        }
    )
    no_generate: bool = field(
        metadata={
            "name_or_flags": ["--no-generate"],
            "default": False,
            "required": False,
            "action": "store_true",
            "help": "Skips the generation of the file structure",
        }
    )
    repositories: List[str] = field(
        metadata={
            "name_or_flags": ["repositories"],
            "nargs": "+",
            "metavar": "repo",
            "help": "The Github repository paths to upload to the index. "
            "Usually in the form <user>/<repo_name>.",
        }
    )

    @classmethod
    def _format(cls: Type[T], parsed: dict[str, Any]) -> T:
        if not parsed["domain"].endswith("/"):
            parsed["domain"] += "/"
        return cls(**parsed)


@dataclass(frozen=True)
class GenerateArgs(Arguments):
    pass


@dataclass(frozen=True)
class YankArgs(Arguments):
    project: str = field(
        metadata={
            "name_or_flags": ["project"],
            "help": "The name of the project to yank.",
        }
    )
    release: str = field(
        metadata={
            "name_or_flags": ["release"],
            "help": "The name of the release to yank.",
        }
    )
    comment: Optional[str] = field(
        metadata={
            "name_or_flags": ["comment"],
            "nargs": "?",
            "help": "The reason for the yanking. [default: None]",
        }
    )



import os
import urllib.parse
from argparse import Action, ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Optional, Sequence, Union


def file_size_str(file_or_size: Union[Path, int]) -> str:
    if isinstance(file_or_size, Path):
        size = file_or_size.stat().st_size
    else:
        size = file_or_size
    suffix = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    i = 0
    while size > 1024 and i < len(suffix) - 1:
        size = size // 1024
        i += 1
    return f"{size} {suffix[i]}"


def delete_path(path: Path) -> None:
    if path.is_dir():
        for child in path.glob("*"):
            delete_path(child)
        path.rmdir()
    else:
        path.unlink(missing_ok=True)


class EnvironmentDefault(Action):
    """Get values from environment variable."""

    def __init__(
        self,
        env: str,
        required: bool = True,
        default: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        default = os.environ.get(env, default)
        self.env = env
        if default:
            required = False
        super().__init__(default=default, required=required, **kwargs)

    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[str] = None,
    ) -> None:
        setattr(namespace, self.dest, values)


def parse_url(url: str) -> str:
    return urllib.parse.urlparse(url).geturl()



from setuptools import setup

setup(
    install_requires=[
        "colorama >= 0.4.4",
        "packaging >= 21.3",
        "requests >= 2.27.1",
        "pkginfo >= 1.8.1",
        "rfc3986 >= 2.0.0",
        "trove-classifiers >= 2022.1.6",
        'readme_renderer >= 32.0',
        'cmarkgfm >= 0.7.0',
    ],
    tests_require=[
        "tox >= 3.24.5",
    ],
)



import re
from dataclasses import dataclass, field
from datetime import datetime
from importlib.resources import path
from typing import Final, Optional

import packaging.utils

import warehub
from warehub.database import Table


class Directory:
    FILES: Final[str] = "files"
    PROJECT: Final[str] = "project"
    SIMPLE: Final[str] = "simple"
    PYPI: Final[str] = "pypi"

    LIST: Final[set[str]] = {FILES, PROJECT, SIMPLE, PYPI}


class Template:
    # STYLE: Final[str] = read_text(warehub.__title__, "style.css")
    # HOMEPAGE: Final[str] = read_text(warehub.__title__, "homepage.html")
    with path(warehub.__title__, "templates") as path:
        HOMEPAGE: Final[str] = (path / "homepage.html").read_text()
        RELEASE: Final[str] = (path / "release.html").read_text()
        SIMPLE: Final[str] = (path / "simple.html").read_text()
        STYLE: Final[str] = (path / "style.css").read_text()


@dataclass
class Project(Table):
    name: str
    created: datetime = field(default_factory=datetime.now)
    documentation: Optional[str] = None
    total_size: int = 0

    def __repr__(self) -> str:
        return f"Project(name={self.name}, created={self.created})"

    @property
    def normalized_name(self):
        return packaging.utils.canonicalize_name(self.name)


@dataclass
class Release(Table):
    project_id: int
    version: str
    created: datetime = field(default_factory=datetime.now)
    author: Optional[str] = None
    author_email: Optional[str] = None
    maintainer: Optional[str] = None
    maintainer_email: Optional[str] = None
    summary: Optional[str] = None
    description: dict[str, str] = field(default_factory=dict)
    keywords: Optional[str] = None
    classifiers: list[str] = field(default_factory=list)
    license: Optional[str] = None
    platform: Optional[str] = None
    home_page: Optional[str] = None
    download_url: Optional[str] = None
    requires_python: Optional[str] = None
    dependencies: dict[str, list[str]] = field(default_factory=dict)
    project_urls: list[str] = field(default_factory=list)
    uploader: Optional[str] = None  # User that created the issue
    uploaded_via: Optional[str] = None
    yanked: bool = False
    yanked_reason: Optional[str] = None

    @property
    def is_pre_release(self):
        return re.match(rf"(a|b|rc)(0|[1-9][0-9]*)", self.version) is not None

    @property
    def urls(self):
        _urls = {}

        if self.home_page:
            _urls["Homepage"] = self.home_page
        if self.download_url:
            _urls["Download"] = self.download_url

        for url_spec in self.project_urls:
            name, _, url = url_spec.partition(",")
            name = name.strip()
            url = url.strip()
            if name and url:
                _urls[name] = url

        return _urls

    # TODO
    # @property
    # def github_repo_info_url(self):
    #     for url in self.urls.values():
    #         parsed = urlparse(url)
    #         segments = parsed.path.strip("/").split("/")
    #         if parsed.netloc in {"github.com", "www.github.com"} and len(segments) >= 2:
    #             user_name, repo_name = segments[:2]
    #             return f"https://api.github.com/repos/{user_name}/{repo_name}"

    @property
    def has_meta(self):
        return any(
            (
                self.license,
                self.keywords,
                self.author,
                self.author_email,
                self.maintainer,
                self.maintainer_email,
                self.requires_python,
            )
        )


@dataclass
class File(Table):
    release_id: int
    name: str
    python_version: Optional[str] = None
    package_type: Optional[str] = None
    comment_text: Optional[str] = None
    size: int = -1
    has_signature: bool = False
    md5_digest: Optional[str] = None
    sha256_digest: Optional[str] = None
    blake2_256_digest: Optional[str] = None
    upload_time: datetime = field(default_factory=datetime.now)
    uploaded_via: Optional[str] = None

    @property
    def pgp_name(self):
        return self.name + ".asc"


@dataclass
class FileName(Table):
    name: str
